<?php

namespace Drupal\spectra\Spectra\Plugin;

use Drupal\Core\Plugin\PluginBase;
use Drupal\Component\Utility\Xss;
use Drupal\spectra\SpectraPluginInterface;
use Symfony\Component\HttpFoundation\Request;

use Drupal\core\Entity;
use Drupal\spectra\Entity\SpectraData;
use Drupal\spectra\Entity\SpectraStatement;

/**
 *
 * @SpectraPlugin(
 *   id = "default_spectra_plugin",
 *   label = @Translation("DefaultSpectraPlugin"),
 * )
 */
class DefaultSpectraPlugin extends PluginBase implements SpectraPluginInterface {

  /**
   * Provide a description of the sandwich.
   * @return string
   *   A string description of the sandwich.
   */
  public function description()
  {
    return $this->t('Default Spectra API Plugin');
  }

  /**
   * {@inheritdoc}
   *
   * Smple helper function to determine whether aray is associative
   */
  public static function is_assoc(array $array)
  {
    return (array_values($array) !== $array);
  }

  /**
   * {@inheritdoc}
   */
  public function handlePostRequest(Request $request) {
    $content = json_decode($request->getContent(), TRUE);
    $objs = array();
    // Create the referenced components
    if (isset($content['actor'])) {
      $objs['actor'] = $this->post_spectra_statement_entity($content['actor'], 'SpectraActor');
    }
    if (isset($content['action'])) {
      $objs['action'] = $this->post_spectra_statement_entity($content['action'], 'SpectraAction');
    }
    if (isset($content['object'])) {
      $objs['object'] = $this->post_spectra_statement_entity($content['object'], 'SpectraObject');
    }
    if (isset($content['context'])) {
      $objs['context'] = $this->post_spectra_statement_entity($content['context'], 'SpectraContext');
    }

    // Create the statement
    $objs['statement'] = $this->post_spectra_statement($objs, $content);

    // Finally, handle the data, as it references the statement, not the other way around
    if (isset($content['data'])) {
      $objs['data'] = $this->post_spectra_data($content['data'], $objs['statement']);
    }

    $ret = [];
    foreach ($objs as $key => $obj) {
      if ($key === 'data') {
        $ret[$key] = array();
        foreach ($obj as $data) {
          $ret[$key][] = $data->uuid();
        }
      }
      else {
        $ret[$key] = $obj->uuid();
      }
    }
    return $ret;
  }

  /**
   * {@inheritdoc}
   */
  public function handleGetRequest($request) {
    $content = $request->query->all();
    if (isset($content['_format'])) {
      unset($content['_format']);
    }
    // Get the type
    if (isset($content['type'])) {
      $type = $content['type'];
      unset($content['type']);
    }
    else {
      $type = 'spectra_statement';
    }

    $search = array();
    foreach ($content as $key => $value) {
      if (strpos($key, '_op') !== FALSE) {
        $value = str_replace('_op', '', $key);
        $search[Xss::filter($key)]['op'] = Xss::filter($value);
      }
      else {
        $search[Xss::filter($key)]['value'] = Xss::filter($value);
      }
    }
    $entities = $this->get_spectra_entities($search, $type);
    $ret = $entities;


    return $ret;
  }

  /**
   * @inheritdoc
   *
   * Gets a Spectra Entity from the database
   *
   * @param (array) search:
   *   Array of conditions with keys as fields to query, and values as associative arrays with a "value" and "op" field.
   *   The "value" field is the search value, and "op" will be a search method such as '>', 'CONTAINS' and so on
   *
   * @param (string) type:
   *   The type of Spectra entity, in a "machine_name" format.
   *
   * @see \Drupal::entityQuery()
   */
  public function get_spectra_entities($search, $type = 'spectra_statement') {
    $query = \Drupal::entityQuery($type);
    foreach ($search as $field => $opts) {
      if (isset($opts['value']) && isset($opts['op'])) {
        $query->condition($field, $opts['value'], $opts['op']);
      }
      else if (isset($opts['value'])) {
        $query->condition($field, $opts['value']);
      }
    }
    $ids = $query->execute();
    return $ids;
  }

  /**
   * @inheritdoc
   *
   * Posts Spectra entities other than Statements and Data into the database
   *
   * @see post_spectra_data
   * @see post_spectra_statement
   */
  public function post_spectra_statement_entity($data, $type) {
    $t = '\\Drupal\\spectra\\Entity\\' . $type;
    $short = $t::getSpectraEntityType('short');

    $search = array();
    if (is_string($data)) {
      $search[$short . '_name']['value'] = $data;
    }
    else if (is_array($data)) {
      foreach($data as $key => $value) {
        strpos($key,'data') === FALSE ? $search[$key]['value'] = $value: NULL;
      }
    }

    // If we have a pre-existing entity, use that. Otherwise, create a new entity

    $ids = $this->get_spectra_entities($search, $t::getSpectraEntityType());

    if ($ids) {
      $id = array_shift($ids);
      $entity = $t::load($id);

      return $entity;
    }
    else {
      if (is_string($data)) {
        $short = $t::getSpectraEntityType('short');
        $new_entity = $t::create(
          array(
            $short . '_type' => 'default',
            $short . '_name' => $data,
            $short . '_data' => json_encode(array()),
          )
        );
        $new_entity->save();
        return $new_entity;
      }
      else if (is_array($data)) {
        // JSON encode the data component
        if (isset($data[$short . '_data'])) {
          $data[$short . '_data'] = json_encode($data[$short . '_data']);
        }
        elseif (isset($data['data'])) {
          $data[$short . '_data'] = json_encode($data['data']);
          unset($data['data']);
        }
        $new_entity = $t::create($data);
        $new_entity->save();
        return $new_entity;
      }
    }
  }

  /**
   * @inheritdoc
   */
  public function post_spectra_data($data, $statement) {
    $d = array();
    if (is_array($data)) {
      // If the array is associative, then we assume it is a single item.
      if (!self::is_assoc($data)) {
        for($i = 0; $i < count($data); $i++) {
          $di['statement_id'] = $statement->id();
          // Get the data type
          if (isset($data[$i]['data_type'])) {
            $di['data_type'] = $data[$i]['data_type'];
          }
          elseif (isset($data[$i]['type'])) {
            $di['data_type'] = $data[$i]['type'];
            unset($data[$i]['type']);
          }
          else {
            $di['data_type'] = 'default';
          }
          // Get the data
          if (isset($data[$i]['data_data'])) {
            $di['data_data'] = json_encode($data[$i]['data_data']);
          }
          elseif (isset($data[$i]['data'])) {
            $di['data_data'] = json_encode($data[$i]['data']);
          }
          else {
            $di['data_data'] = json_encode($data[$i]);
          }
          $d[] = $di;
        }
      }
      else {
        $di['statement_id'] = $statement->id();
        // Get the data type
        if (isset($data['data_type'])) {
          $di['data_type'] = $data['data_type'];
        }
        elseif (isset($data['type'])) {
          $di['data_type'] = $data['type'];
          unset($data['type']);
        }
        else {
          $di['data_type'] = 'default';
        }
        // Get the data
        if (isset($data['data_data'])) {
          $di['data_data'] = json_encode($data['data_data']);
        }
        elseif (isset($data['data'])) {
          $di['data_data'] = json_encode($data['data']);
          unset($data['type']);
        }
        else {
          $di['data_data'] = json_encode($data);
        }
        $d[] = $di;
      }
    }
    else {
      $di['statement_id'] = $statement->id();
      $di['data_type'] = 'default';
      $di['data_data'] = json_encode($data);
      $d[] = $di;
    }

    $ret = array();
    for($i = 0; $i < count($d); $i++) {
      $data_entity = SpectraData::create($d[$i]);
      $data_entity->save();
      $ret[] = $data_entity;
    }
    return $ret;
  }

  /**
   * @inheritdoc
   */
  public function post_spectra_statement($objs, $content) {
    // Create the entity, then assign
    $s = array();
    $s['statement_time'] = isset($content['time']) ? $content['time']: time();
    $s['statement_type'] = isset($content['type']) ? $content['type'] : 'default';

    foreach ($objs as $key => $obj) {
      $s[$key . '_id'] = $obj->id();
    }

    $statement = SpectraStatement::create($s);
    $statement->save();

    return $statement;
  }

}